import { randomID } from '../../util/helpers.js';
import { Box } from './box.js';

const updateValuesOnEventTypes = ['keyup', 'mouseup'];

export class FormElement extends Box {
  constructor(name, form, opts) {
    const parentEl = opts.parentEl ?? form.el;
    super(name, parentEl, opts);
    this.form = form;
    this.id = opts.id ?? name;
    this.includeInOutput = opts.includeInOutput ?? true;
    const { cb } = opts;
    if (cb) {
      updateValuesOnEventTypes.forEach((eventType) => this.el.addEventListener(eventType, () => {
        cb(this, { initializing: false });
      }));
      cb(this, { initializing: true });
    }
  }
}

export class Button extends FormElement {
  constructor(name, form, opts) {
    super(name, form, opts);
    this.button = document.createElement('button');
    this.button.setAttribute('type', 'button');
    this.button.innerHTML = name;
    this.button.disabled = !!opts.disabled;
    this.el.appendChild(this.button);
  }
}

export class TextField extends FormElement {
  constructor(name, form, opts) {
    super(name, form, opts);
    this.label = document.createElement('label');
    this.labelDiv = document.createElement('div');
    this.label.appendChild(this.labelDiv);
    this.labelDiv.innerHTML = name;
    this.input = document.createElement('input');
    this.input.disabled = !!opts.disabled;
    this.input.defaultValue = opts.defaultValue || '';
    this.label.appendChild(this.input);
    this.el.appendChild(this.label);
  }

  get value() {
    return this.input?.value || null;
  }
}

export class TextArea extends FormElement { }

export class SubFormArray extends FormElement {
  constructor(name, form, opts) {
    super(name, form, opts);
    this.subForms = [];
  }

  get value() {
    return this.subForms.map((subForm) => subForm.value);
  }

  remove(subForm) {
    const idx = this.subForms.findIndex((s) => s === subForm);
    this.subForms.splice(idx, 1);
    subForm.el.remove();
  }
}

export class SubForm extends FormElement {
  constructor(name, form, opts) {
    const parentEl = opts.subFormArray ? opts.subFormArray.el : form.el;
    const subForm = form.document.form({ name, parentEl }).lastElement;
    super(name, form, { ...opts, parentEl });
    this.subForm = subForm;
    if (opts.subFormArray) {
      opts.subFormArray.subForms.push(this);
      this.includeInOutput = false;
    }
  }

  get value() {
    return this.subForm.value;
  }
}

export class Form extends Box {
  constructor(document, opts = {}) {
    super(opts.name, opts.parentEl || document.el, opts);
    this.document = document;
    this.items = [];
    this.id = opts.id ?? `form_${randomID()}`;
  }

  button(opts) {
    this.items.push(new Button(opts.name, this, opts));
    return this;
  }

  textField(opts) {
    this.items.push(new TextField(opts.name, this, opts));
    return this;
  }

  textArea(opts) {
    this.items.push(new TextArea(opts.name, this, opts));
    return this;
  }

  subForm(opts) {
    this.items.push(new SubForm(opts.name, this, opts));
    return this;
  }

  subFormArray(opts) {
    this.items.push(new SubFormArray(opts.name, this, opts));
    return this;
  }

  get lastItem() {
    return this.items[this.items.length - 1];
  }

  get value() {
    return this.items.reduce((result, { id, value, includeInOutput }) => {
      if (includeInOutput && value !== undefined) {
        result[id] = value;
      }
      return result;
    }, {});
  }
}
